from .browser import Browser


class FirefoxHeadless(Browser):

    def __init__(self):
        raise NotImplementedError(
            "Could not make screenshot work on Firefox headless yet ...\n"
            "See https://bugzilla.mozilla.org/show_bug.cgi?id=1715450"
        )

    @property
    def executable(self):
        pass

    @executable.setter
    def executable(self, value):
        pass

    def render(self, **kwargs):
        pass

    def __enter__(self):
        pass

    def __exit__(self, *exc):
        pass



import pytest
from jaraco import path

from setuptools.glob import glob


@pytest.mark.parametrize(
    ('tree', 'pattern', 'matches'),
    (
        ('', b'', []),
        ('', '', []),
        (
            """
     appveyor.yml
     CHANGES.rst
     LICENSE
     MANIFEST.in
     pyproject.toml
     README.rst
     setup.cfg
     setup.py
     """,
            '*.rst',
            ('CHANGES.rst', 'README.rst'),
        ),
        (
            """
     appveyor.yml
     CHANGES.rst
     LICENSE
     MANIFEST.in
     pyproject.toml
     README.rst
     setup.cfg
     setup.py
     """,
            b'*.rst',
            (b'CHANGES.rst', b'README.rst'),
        ),
    ),
)
def test_glob(monkeypatch, tmpdir, tree, pattern, matches):
    monkeypatch.chdir(tmpdir)
    path.build({name: '' for name in tree.split()})
    assert list(sorted(glob(pattern))) == list(sorted(matches))



"""
Main file of the html2image package.

html2image is a package acting as a wrapper around the
headless mode of existing web browsers to generate images
from URLs and from HTML+CSS strings or files.

For feedback, usage and to learn more, see https://github.com/vgalin/html2image
"""

import os
import shutil

from textwrap import dedent

from html2image.browsers import chrome, chrome_cdp, edge  # , firefox, firefox_cdp
from html2image.browsers.browser import Browser, CDPBrowser


browser_map = {
    'chrome': chrome.ChromeHeadless,
    'chromium': chrome.ChromeHeadless,
    'google-chrome': chrome.ChromeHeadless,
    'google-chrome-stable': chrome.ChromeHeadless,
    'googlechrome': chrome.ChromeHeadless,
    'edge': edge.EdgeHeadless,
    'chrome-cdp': chrome_cdp.ChromeCDP,
    'chromium-cdp': chrome_cdp.ChromeCDP,
    # 'firefox': firefox.FirefoxHeadless,
    # 'mozilla-firefox': firefox.FirefoxHeadless,
    # 'firefox-cdp': firefox_cdp.FirefoxCDP,
}


class Html2Image():
    """
        Allows the generation of images from
        URLs and HTML/CSS files or strings.

        Parameters
        ----------
        - `browser`: str , optional
            + Type of the browser that will be used to take screenshots.
            + Default is Chrome.

        - `browser_executable` : str, optional
            + Path to a browser executable.

        - `output_path` : str, optional
            + Path to a directory in which the taken screenshots will be saved.
            + Default is the current working directory.

        - `size` : (int, int), optional
            + Size of the screenshots.
            + Default is (1920, 1080).

        - `temp_path` : str, optional
            + Path to a directory that will be used to store temporary files.

        - `keep_temp_files` : bool, optional
            + If True, will not automatically remove temporary files created.

        - `custom_flags`: list of str or str, optional
            + Additional custom flags for the headless browser.

        Raises
        ------
        - `FileNotFoundError`
            + If an executable of the browser specified in the `browser`
            parameter was not found.
    """

    def __init__(
        self,
        browser='chrome',
        browser_executable=None,
        browser_cdp_port=None,
        output_path=os.getcwd(),
        size=(1920, 1080),
        temp_path=None,
        keep_temp_files=False,
        custom_flags=None,
        disable_logging=False,
    ):

        if browser.lower() not in browser_map:
            raise ValueError(
                f'"{browser}" is not a browser known by HTML2Image.'
            )

        self.output_path = output_path
        self.size = size
        self.temp_path = temp_path
        self.keep_temp_files = keep_temp_files
        self.browser: Browser = None

        browser_class = browser_map[browser.lower()]

        if isinstance(browser_class, CDPBrowser):
            self.browser = browser_class(
                executable=browser_executable,
                flags=custom_flags,
                cdp_port=browser_cdp_port,
                disable_logging=disable_logging,
            )
        else:
            self.browser = browser_class(
                executable=browser_executable,
                flags=custom_flags,
                disable_logging=disable_logging,
            )

    @property
    def temp_path(self):
        return self._temp_path

    @temp_path.setter
    def temp_path(self, value):
        if value is None:
            temp_dir = os.environ['TMP'] if os.name == 'nt' else '/tmp'
            temp_dir = os.path.join(temp_dir, 'html2image')
        else:
            temp_dir = value

        # create the directory if it does not exist
        os.makedirs(temp_dir, exist_ok=True)

        self._temp_path = temp_dir

    @property
    def output_path(self):
        return self._output_path

    @output_path.setter
    def output_path(self, value):
        # output_path should always be an absolute path
        value = os.path.abspath(value)

        # create the directory if it does not exist
        os.makedirs(value, exist_ok=True)

        self._output_path = value

    def load_str(self, content, as_filename):
        """
        Loads a string containing HTML or CSS so that html2image can use it
        later to take a screenshot.

        Behind the scenes the string (that can really contain anything) is
        written into a file that is saved in the directory defined in the
        `temp_path` attribute.

        Parameters
        ----------
        - `content`: str
            + HTML/CSS formatted text.

        - `as_filename`: str
            + Filename as which the given string will be saved.

        """
        with open(os.path.join(self.temp_path, as_filename), 'wb') as f:
            f.write(content.encode('utf-8'))

    def load_file(self, src, as_filename=None):
        """ 'Loads' a file that html2image can use later to take a screenshot.

        Behind the scenes, the file found at `src` is:
        -   eventually renamed, if the `as_filename` parameter is specified;
        -   then sent to the directory defined in the  `temp_path` attribute.

        Parameters
        ----------
        - `src`: str
            + Path to the file to load.

        - `as_filename`: str
            + Filename as which the given file will renamed.
            + If None or not specified, the given file will keep
            its original name.
        """
        if as_filename is None:
            as_filename = os.path.basename(src)

        dest = os.path.join(self.temp_path, as_filename)
        shutil.copyfile(src, dest)

    def _remove_temp_file(self, filename):
        """ Removes a file in the tmp directory.

        This function is used after a temporary file is created in order to
        load an HTML string.
        This prevents the temp directory to end up bloated by temp files.

        Parameters
        ----------
        - `filename`: str
            + Filename of the file to be removed
            + (path is the temp_path directory)
        """
        os.remove(os.path.join(self.temp_path, filename))

    def screenshot_loaded_file(
        self, file, output_file='screenshot.png', size=None
    ):
        """ Takes a screenshot of a *previously loaded* file or string.

        Parameters
        ----------
        - `file`: str
            + HTML file that will be screenshotted.

        - `output_file`: str
            + Name as which the screenshot will be saved.
            + File extension (e.g. .png) has to be included.
            + Default is screenshot.png

        - `size`: (int, int), optional
            + Size of the screenshot that will be taken when the
            method is called.
        """

        file = os.path.join(self.temp_path, file)

        if os.path.dirname(output_file) != '':
            raise ValueError(
                "the output_file parameter should be a filename "
                "and not a path.\nChange the output path by "
                "modifying the output_path attribute."
            )

        self.browser.screenshot(
            output_path=self.output_path,
            output_file=output_file,
            input=file,
            size=size,
        )

    def screenshot_url(self, url, output_file='screenshot.png', size=None):
        """ Takes a screenshot of a given URL.

        The given URL should be well formed or it may result in undefined
        behaviors when an headless browser will open it.
        Please do include the protocol in the URL (http, https).
        E.g. url = 'https://www.python.org/'

        Parameters
        ----------
        - `url`: str
            + URL of the page that will be screenshotted.
            + Do not ommit the protocol.

        - `output_file`: str, optional
            + Name as which the screenshot will be saved.
            + File extension (e.g. .png) has to be included.
            + Default is screenshot.png

        - `size`: (int, int), optional
            + Size of the screenshot that will be taken when the
            + method is called.
        """

        if os.path.dirname(output_file) != '':
            raise ValueError(
                "the output_file parameter should be a filename "
                "and not a path.\nChange the output path by "
                "modifying the output_path attribute."
            )

        self.browser.screenshot(
            output_path=self.output_path,
            output_file=output_file,
            input=url,
            size=size
        )

    @staticmethod
    def _extend_save_as_param(save_as, desired_length):
        """ Extend the save_as parameter of the `screenshot()` method.

        So we do not run out of filenames.

        Parameters
        ----------
        - `save_as`: list
            + List of filenames
        - `desired_length`: int
            + Minimum desired length of the ouput

        Returns
        -------
        - list
            + `save_as` extended to `desired_length`

        Examples
        --------
        >>> _extend_save_as_param(['a.png', 'b.png'], 2)
        ['a.png', 'b.png']

        >>> _extend_save_as_param(['a.png', 'b.png'], 4)
        ['a.png', 'b_0.png', 'b_1.png', 'b_2.png']

        >>> _extend_save_as_param(['a.png', 'b.png'], 0)
        ['a.png', 'b.png']

        >>> _extend_save_as_param(['a.png', 'b.png', None, 65], 2)
        ['a.png', 'b.png']
        """

        # get rid of anything that is not a string
        save_as = [name for name in save_as if isinstance(name, str)]

        if desired_length <= len(save_as):
            return save_as

        missing_name_count = desired_length - len(save_as)
        filename, extention = save_as[-1].split('.')

        # remove last object as it will be replaced
        # from filename.extention to filename_0.extention
        save_as.pop()

        save_as.extend([
            f'{filename}_{i}.{extention}'
            for i in range(missing_name_count + 1)
        ])

        return save_as

    def _extend_size_param(self, sizes, desired_length):
        """ Extend the size parameter of the `screenshot()` method.

        So we do not run out of sizes.
        If the given the `sizes` parameter is an empty list, the list
        will the extended using `self.size`.

        Parameters
        ----------
        - `save_as`: list
            + List of (int, int) tuples
        - `desired_length`: int
            + Minimum desired length of the ouput

        Returns
        -------
        - list
            + `sizes` extended to `desired_length`

        Examples
        --------
        >>> _extend_size_param([(50, 50)], 1)
        [(50, 50)]

        >>> _extend_size_param([(50, 50)], 3)
        [(50, 50), (50, 50), (50, 50)]

        >>> _extend_size_param([(50, 50), (70, 60), (80, 90)], 5)
        [(50, 50), (70, 60), (80, 90), (80, 90), (80, 90)]

        >>> _extend_size_param([], 3)
        [(1920, 1080), (1920, 1080), (1920, 1080)]

        """

        # get rid of anything that is not a string
        sizes = [
            size for size in sizes
            if isinstance(size, tuple) and len(size) == 2
        ]

        if desired_length <= len(sizes):
            return sizes

        # if no size specified, then use default size
        if len(sizes) == 0:
            return [
                self.size
                for i in range(desired_length)
            ]

        missing_size_count = desired_length - len(sizes)
        last_size = sizes[-1]

        sizes.extend([
            last_size
            for i in range(missing_size_count)
        ])

        return sizes

    @staticmethod
    def _prepare_html_string(html_body, css_style_string) -> str:
        """ Creates a basic HTML string from an HTML body and a css string.

        Parameters
        ----------
        - `html_body`: str
        - `css_style_string`: str

        Returns
        -------
        - str
            A combination of `html_body` and `css_style_string` put
            together in an HTML template.

        """

        prepared_html = f"""\
        <html>
        <head>
          <meta charset="UTF-8">
            <style>
                {css_style_string}
            </style>
        </head>

        <body>
            {html_body}
        </body>
        </html>
        """
        return dedent(prepared_html)

    @staticmethod
    def _prepare_css_string(css_file: list[str])-> str:
        """ Creates a basic string fromatted from a list of css files.

        Parameters
        ----------
        - `css_file`: str

        Returns
        -------
        - str
            The concatenated content of each css files.

        """

        css_str = ''
        for css in css_file:
            temp_css_str = ''
            with open(css, "r") as fd:
                temp_css_str = fd.read()
                css_str += temp_css_str + '\n'

        return css_str


    def screenshot(
        self,
        html_str=[],  # html_str: Union[str, list] = [],
        html_file=[],
        css_str=[],
        css_file=[],
        other_file=[],
        url=[],
        save_as='screenshot.png',
        size=[],
    ):
        """ Takes a screenshot using different resources.

        Parameters
        ----------
        - `html_str`: list of str or str
            + HTML string(s) that will be screenshotted.
        - `html_file`: list of str or str
            + Filepath(s) of HTML file(s) that will be screenshotted.
        - `css_str`: list of str or str
            + CSS string(s) that will be "associated" with the given
            + HTML string(s).
        - `css_file`: list of str or str
            + Filepath(s) of CSS file(s). These files serve two purposes:
            +   1. Their content is combined with `css_str` and embedded
            +      directly when screenshotting `html_str` (HTML strings).
            +   2. They are loaded into the temporary directory, making
            +      them available to `html_file` (HTML files) that link to them
            +      (e.g., via `<link rel="stylesheet" href="filename.css">`).
        - `other_file`: list of str or str
            + Filepath(s) of non-HTML file(s) that will be screenshotted.
        - `url`: list of str or str
            + URL(s) of the page(s) that will be screenshotted.
            + Do not ommit the protocol.
        - `save_as`: list of str or str
            + Name(s) as which the screenshot will be saved.
            + File extension (e.g. .png) has to be included.
            + Default value is screenshot.png
        - `size`: list of (int, int) or (int, int) tuple
            + Size(s) of the screenshot(s) that will be taken when the
            + method is called.

        Returns
        -------
        - list of str
            + A list of the file path(s) of the generated image(s)

        Raises
        ------
        - `FileNotFoundError`
        """

        # TODO / NOTE : This does not pose any problem for now but setting
        # mutables (here empty lists) as default arguments of a function
        # can cause unwanted behaviours.

        screenshot_paths = []

        # convert each parameter into list
        # e.g: param=value becomes param=[value]
        html_strings = [html_str] if isinstance(html_str, str) else html_str
        html_files = [html_file] if isinstance(html_file, str) else html_file
        css_strings = [css_str] if isinstance(css_str, str) else css_str
        css_files = [css_file] if isinstance(css_file, str) else css_file
        other_files = [other_file] if isinstance(other_file, str) else other_file
        urls = [url] if isinstance(url, str) else url
        save_as_filenames = [save_as] if isinstance(save_as, str) else save_as
        sizes = [size] if isinstance(size, tuple) else size

        planned_screenshot_count = (
            len(html_strings) + len(html_files) + len(other_files) + len(urls)
        )
        save_as_filenames = Html2Image._extend_save_as_param(
            save_as_filenames,
            planned_screenshot_count,
        )
        sizes = self._extend_size_param(sizes, planned_screenshot_count)

        css_style_string = '\n'.join(css_strings) + '\n'

        if css_files:
            # add content from css_files, regardless of whether css_strings was present
            css_style_string += Html2Image._prepare_css_string(css_files)


        for css in css_files:
            if os.path.isfile(css):
                self.load_file(src=css)
            else:
                raise FileNotFoundError(css)

        for html in html_strings:
            name = save_as_filenames.pop(0)
            current_size = sizes.pop(0)

            base_name, _ = os.path.splitext(name)
            html_filename = base_name + '.html'

            content = Html2Image._prepare_html_string(html, css_style_string)

            self.load_str(content=content, as_filename=html_filename)
            self.screenshot_loaded_file(
                file=html_filename,
                output_file=name,
                size=current_size,
            )
            if not self.keep_temp_files:
                self._remove_temp_file(html_filename)

            screenshot_paths.append(os.path.join(self.output_path, name))

        for screenshot_target in html_files + other_files:

            name = save_as_filenames.pop(0)
            current_size = sizes.pop(0)

            if os.path.isfile(screenshot_target):
                self.load_file(src=screenshot_target)
                self.screenshot_loaded_file(
                    file=os.path.basename(screenshot_target),
                    output_file=name,
                    size=current_size,
                )
                if not self.keep_temp_files:
                    self._remove_temp_file(os.path.basename(screenshot_target))                
            else:
                raise FileNotFoundError(screenshot_target)

            screenshot_paths.append(os.path.join(self.output_path, name))

        for target_url in urls:
            name = save_as_filenames.pop(0)
            current_size = sizes.pop(0)

            self.screenshot_url(
                url=target_url,
                output_file=name,
                size=current_size,
            )
            screenshot_paths.append(os.path.join(self.output_path, name))

        return screenshot_paths

    def __enter__(self):
        self.browser.__enter__()
        return self

    def __exit__(self, *exc):
        self.browser.__exit__(*exc)


if __name__ == '__main__':
    pass



import os
import platform
import shutil
import subprocess

try:
    from winreg import ConnectRegistry, OpenKey, QueryValueEx,\
            HKEY_LOCAL_MACHINE, HKEY_CURRENT_USER, KEY_READ
except ImportError:
    # os is not Windows, and there is no need for winreg
    pass

ENV_VAR_LOOKUP_TOGGLE = 'HTML2IMAGE_TOGGLE_ENV_VAR_LOOKUP'

CHROME_EXECUTABLE_ENV_VAR_CANDIDATES = [
    'HTML2IMAGE_CHROME_BIN',
    'HTML2IMAGE_CHROME_EXE',
    'CHROME_BIN',
    'CHROME_EXE',
]

FIREFOX_EXECUTABLE_ENV_VAR_CANDIDATES = [
    'HTML2IMAGE_FIREFOX_BIN',
    'HTML2IMAGE_FIREFOX_EXE',
    'FIREFOX_BIN',
    'FIREFOX_EXE',
]


def get_command_origin(command):
    ''' Finds the path of a given command (windows only).

    This function is inspired by the `start` command on windows.
    It will search if the given command corresponds :
    - To an executable in the current directory
    - To an executable in the PATH environment variable
    - To an executable mentionned in the registry in the
      HKEY_LOCAL_MACHINE or HKEY_LOCAL_MACHINE hkeys in
      SOFTWARE\\[Wow6432Node\\]Microsoft\\Windows\\CurrentVersion\\App Paths\\

    It also "indirectly" validates a path to a file.

    Parameters
    ----------
    - `command`: str
        + command that will be searched.

    Returns
    -------
    - str or None
        + the path corresponding to `command`
        + None, if no path was found
    '''

    # `start "command"` itself could be used to run the command but we want to
    # assess that the command is a valid one.
    command = command.replace('start ', '')

    # which search in current directory + path
    # and file extention can be ommited
    which_result = shutil.which(command)
    if which_result:
        return which_result

    # search for command in registry

    command = command.replace('.exe', '').strip()

    # Once combined, these hkeys and keys are where the `start`
    # command seach for executable.
    hkeys = [HKEY_LOCAL_MACHINE, HKEY_CURRENT_USER]
    keys = [
        "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\"
        f"App Paths\\{command}.exe",

        "SOFTWARE\\Wow6432Node\\Microsoft\\Windows\\"
        "CurrentVersion\\App Paths\\"
        f"{command}.exe",
    ]

    for hkey in hkeys:
        with ConnectRegistry(None, hkey) as registry:
            for key in keys:
                try:
                    with OpenKey(registry, key, 0, KEY_READ) as k:
                        # if no exception: we found the exe
                        return QueryValueEx(k, '')[0]
                except Exception:
                    # cannot open key, do nothing and proceed to the next one
                    pass
    return None


def find_first_defined_env_var(env_var_list, toggle):
    '''
    Returns the value of the first defined environment variable
    encountered in the `env_var_list` list, but only if the
    the `toggle` environment variableif defined.

    Parameters
    ----------
    - `env_var_list`: list[str]
        + list of environment variable names
    - `toggle`: str
        + environment variable name


    '''
    if toggle in os.environ:
        for env_var in env_var_list:
            value = os.environ.get(env_var)
            if value:
                return value
    return None


def find_chrome(user_given_executable=None):
    """ Finds a Chrome executable.

    Search Chrome on a given path. If no path given,
    try to find Chrome or Chromium-browser on a Windows or Unix system.

    Parameters
    ----------
    - `user_given_executable`: str (optional)
        + A filepath leading to a Chrome/ Chromium executable
        + Or a filename found in the current working directory
        + Or a keyword that executes Chrome/ Chromium, ex:
            - 'chromium' on linux systems
            - 'chrome' on windows (if typing `start chrome` in a cmd works)

    Raises
    ------
    - `FileNotFoundError`
        + If a suitable chrome executable could not be found.

    Returns
    -------
    - str
        + Path of the chrome executable on the current machine.
    """

    # try to find a chrome bin/exe in ENV
    path_from_env = find_first_defined_env_var(
        env_var_list=CHROME_EXECUTABLE_ENV_VAR_CANDIDATES,
        toggle=ENV_VAR_LOOKUP_TOGGLE
    )

    if path_from_env:
        print(
            f'Found a potential chrome executable in the {path_from_env} '
            f'environment variable:\n{path_from_env}\n'
        )
        return path_from_env

    # if an executable is given, try to use it
    if user_given_executable is not None:

        # On Windows, we cannot "safely" validate that user_given_executable
        # seems to be a chrome executable, as we cannot run it with
        # the --version flag.
        # https://bugs.chromium.org/p/chromium/issues/detail?id=158372
        #
        # We thus do the "bare minimum" and check if user_given_executable
        # is a file, a filepath, or corresponds to a keyword that can be used
        # with the start command, like so: `start user_given_executable`
        if platform.system() == 'Windows':
            command_origin = get_command_origin(user_given_executable)
            if command_origin:
                return command_origin

            # cannot validate user_given_executable
            raise FileNotFoundError()

        # On a non-Windows OS, we can validate in a basic way that
        # user_given_executable leads to a Chrome / Chromium executable,
        # or is a command, using the --version flag
        else:
            try:
                if 'chrom' in subprocess.check_output(
                    [user_given_executable, '--version']
                ).decode('utf-8').lower():
                    return user_given_executable
            except Exception:
                pass

        # We got a user_given_executable but couldn't validate it
        raise FileNotFoundError(
            'Failed to find a seemingly valid chrome executable '
            'in the given path.'
        )

    # Executable not in ENV or given by the user, try to find it
    # Search for executable on a Windows OS
    if platform.system() == 'Windows':
        prefixes = [
            os.getenv('PROGRAMFILES(X86)'),
            os.getenv('PROGRAMFILES'),
            os.getenv('LOCALAPPDATA'),
        ]

        suffix = "Google\\Chrome\\Application\\chrome.exe"

        for prefix in prefixes:
            path_candidate = os.path.join(prefix, suffix)
            if os.path.isfile(path_candidate):
                return path_candidate

    # Search for executable on a Linux OS
    elif platform.system() == "Linux":

        chrome_commands = [
            'chromium',
            'chromium-browser',
            'chrome',
            'google-chrome',
            'google-chrome-stable'
        ]

        for chrome_command in chrome_commands:
            if shutil.which(chrome_command):
                # check the --version for "chrom" ?
                return chrome_command

        # snap seems to be a special case?
        # see https://stackoverflow.com/q/63375327/12182226

        try:
            version_result = subprocess.check_output(
                ["chromium-browser", "--version"]
            )
            if 'snap' in str(version_result):
                chrome_snap = (
                    '/snap/chromium/current/usr/lib/chromium-browser/chrome'
                )
                if os.path.isfile(chrome_snap):
                    return chrome_snap
        except Exception:
            pass

    # Search for executable on MacOS
    elif platform.system() == "Darwin":
        # MacOS system
        chrome_app = (
            '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
        )

        try:
            version_result = subprocess.check_output(
                [chrome_app, "--version"]
            )
            if "Google Chrome" in str(version_result):
                return chrome_app
        except Exception:
            pass

    # Couldn't find an executable (or OS not in Windows, Linux or Mac)
    raise FileNotFoundError(
        'Could not find a Chrome executable on this '
        'machine, please specify it yourself.'
    )

def find_firefox(user_given_executable=None):
    """ Finds a Firefox executable.

    Search Firefox on a given path. If no path given,
    try to find Firefox on a Windows or Unix system.

    Parameters
    ----------
    - `user_given_executable`: str (optional)
        + A filepath leading to a Firefox executable
        + Or a filename found in the current working directory
        + Or a keyword that executes Firefox, ex:
            - 'firefox' on linux systems
            - 'firefox' on windows (if typing `start firefox` in a cmd works)

    Raises
    ------
    - `FileNotFoundError`
        + If a suitable Firefox executable could not be found.

    Returns
    -------
    - str
        + Path of the Firefox executable on the current machine.
    """

    # try to find a firefox bin/exe in ENV
    path_from_env = find_first_defined_env_var(
        env_var_list=FIREFOX_EXECUTABLE_ENV_VAR_CANDIDATES,
        toggle=ENV_VAR_LOOKUP_TOGGLE
    )

    if path_from_env:
        print(
            f'Found a potential Firefox executable in the {path_from_env} '
            f'environment variable:\n{path_from_env}\n'
        )
        return path_from_env

    # if an executable is given, try to use it
    if user_given_executable is not None:

        if platform.system() == 'Windows':
            user_given_executable = get_command_origin(user_given_executable)

        try:
            version_output = subprocess.check_output(
                [user_given_executable, '--version']
            ).decode('utf-8').lower()

            if 'Mozilla Firefox' in version_output:
                return user_given_executable
            else:
                print(
                    'Could not validate Firefox executable',
                    '(--version does not contains "Mozilla Firefox").'
                )
        except Exception:
            pass

        # We got a user_given_executable but couldn't validate it
        raise FileNotFoundError(
            'Failed to find a seemingly valid Firefox executable '
            'in the given path.'
        )

    # Executable not in ENV or given by the user, try to find it
    # Search for executable on a Windows OS
    if platform.system() == 'Windows':
        prefixes = [
            os.getenv('PROGRAMFILES(X86)'),
            os.getenv('PROGRAMFILES'),
            os.getenv('LOCALAPPDATA'),
        ]

        suffix = 'Mozilla Firefox\\firefox.exe'

        for prefix in prefixes:
            path_candidate = os.path.join(prefix, suffix)
            if os.path.isfile(path_candidate):
                return path_candidate

    # Search for executable on a Linux OS
    elif platform.system() == 'Linux':
        if shutil.which('firefox'):
            return 'firefox'

    # Search for executable on MacOS
    elif platform.system() == 'Darwin':
        # MacOS system

        # TODO : check if this is the right path
        firefox_app = (
            '/Applications/Firefox.app/Contents/MacOS/firefox'  # ?
        )

        try:
            version_result = subprocess.check_output(
                [firefox_app, '--version']
            )
            if 'Mozilla Firefox' in str(version_result):
                return firefox_app
        except Exception:
            pass

    # Couldn't find an executable (or OS not in Windows, Linux or Mac)
    raise FileNotFoundError(
        'Could not find a Chrome executable on this '
        'machine, please specify it yourself.'
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division, print_function

def decode(data):
    """
    Decodes a PackBit encoded data.
    """
    data = bytearray(data) # <- python 2/3 compatibility fix
    result = bytearray()
    pos = 0
    while pos < len(data):
        header_byte = data[pos]
        if header_byte > 127:
            header_byte -= 256
        pos += 1

        if 0 <= header_byte <= 127:
            result.extend(data[pos:pos+header_byte+1])
            pos += header_byte+1
        elif header_byte == -128:
            pass
        else:
            result.extend([data[pos]] * (1 - header_byte))
            pos += 1

    return bytes(result)


def encode(data):
    """
    Encodes data using PackBits encoding.
    """
    if len(data) == 0:
        return data

    if len(data) == 1:
        return b'\x00' + data

    data = bytearray(data)

    result = bytearray()
    buf = bytearray()
    pos = 0
    repeat_count = 0
    MAX_LENGTH = 127

    # we can safely start with RAW as empty RAW sequences
    # are handled by finish_raw()
    state = 'RAW'

    def finish_raw():
        if len(buf) == 0:
            return
        result.append(len(buf)-1)
        result.extend(buf)
        buf[:] = bytearray()

    def finish_rle():
        result.append(256-(repeat_count - 1))
        result.append(data[pos])

    while  pos < len(data)-1:
        current_byte = data[pos]

        if data[pos] == data[pos+1]:
            if state == 'RAW':
                # end of RAW data
                finish_raw()
                state = 'RLE'
                repeat_count = 1
            elif state == 'RLE':
                if repeat_count == MAX_LENGTH:
                    # restart the encoding
                    finish_rle()
                    repeat_count = 0
                # move to next byte
                repeat_count += 1

        else:
            if state == 'RLE':
                repeat_count += 1
                finish_rle()
                state = 'RAW'
                repeat_count = 0
            elif state == 'RAW':
                if len(buf) == MAX_LENGTH:
                    # restart the encoding
                    finish_raw()

                buf.append(current_byte)

        pos += 1

    if state == 'RAW':
        buf.append(data[pos])
        finish_raw()
    else:
        repeat_count += 1
        finish_rle()

    return bytes(result)



import importlib.resources as pkg_resources
import subprocess
import os

from html2image import Html2Image
import jinja2

from tasks import Task


def print_task(task: Task):
    hti = Html2Image(size=(1109, 696))
    with pkg_resources.path("asup", "template.html") as search_path:
        templateLoader = jinja2.FileSystemLoader(searchpath=search_path)
        templateEnv = jinja2.Environment(loader=templateLoader)

        TEMPLATE_FILE = "template.html"
        template = templateEnv.get_template(TEMPLATE_FILE)

        templateVars = {
            "header": task.name,
            "content": task.description,
            "footer": "•" * task.priority,
        }
        cwd = os.getcwd()

        outputText = template.render(templateVars)
        with open(cwd + "out.html", "w+") as f:
            f.write(outputText)

    hti.screenshot(
        url="file://" + cwd + "/out.html",
        save_as="asup.png",
    )
    subprocess.run(
        [
            "brother_ql",
            "print",
            "-l",
            "62",
            "--red",
            "-r",
            "90",
            "asup.png",
        ],
        check=True,
    )
    os.remove("asup.png")
    os.remove("out.html")


if __name__ == "__main__":
    task = Task(
        name="Example Task",
        description="This is an example task description. 123",
        completed=False,
        priority=3,
    )
    print_task(task)



from .browser import Browser

import os
import subprocess

class ChromiumHeadless(Browser):
    """
        Chrome/Chromium browser wrapper.

        Parameters
        ----------
        - `executable` : str, optional
            + Path to a chrome executable.
        - `flags` : list of str
            + Flags to be used by the headless browser.
            + Default flags are :
                - '--default-background-color=00000000'
                - '--hide-scrollbars'
        - `print_command` : bool
            + Whether or not to print the command used to take a screenshot.
        - `disable_logging` : bool
            + Whether or not to disable Chrome's output.
        - `use_new_headless` : bool, optional
            + Whether or not to use the new headless mode.
            + By default, the old headless mode is used.
            + You can also keep the original behavior to backward compatibility by setting this to `None`.
    """

    def __init__(self, executable=None, flags=None, print_command=False, disable_logging=False, use_new_headless=None,):
        self.executable = executable
        if not flags:
            self.flags = [
                '--default-background-color=00000000',
                '--hide-scrollbars',
            ]
        else:
            self.flags = [flags] if isinstance(flags, str) else flags

        self.print_command = print_command
        self.disable_logging = disable_logging
        self.use_new_headless = use_new_headless

    def screenshot(
        self,
        input,
        output_path,
        output_file='screenshot.png',
        size=(1920, 1080),
    ):
        """ Calls Chrome or Chromium headless to take a screenshot.

            Parameters
            ----------
            - `output_file`: str
                + Name as which the screenshot will be saved.
                + File extension (e.g. .png) has to be included.
                + Default is screenshot.png
            - `input`: str
                + File or url that will be screenshotted.
                + Cannot be None
            - `size`: (int, int), optional
                + Two values representing the window size of the headless
                + browser and by extention, the screenshot size.
                + These two values must be greater than 0.
            Raises
            ------
            - `ValueError`
                + If the value of `size` is incorrect.
                + If `input` is empty.
        """

        if not input:
            raise ValueError('The `input` parameter is empty.')

        if size[0] < 1 or size[1] < 1:
            raise ValueError(
                f'Could not screenshot "{output_file}" '
                f'with a size of {size}:\n'
                'A valid size consists of two integers greater than 0.'
            )

        # command used to launch chrome in
        # headless mode and take a screenshot
        headless_mode = '--headless'
        if self.use_new_headless is not None:
            headless_mode += '=new' if self.use_new_headless else '=old'

        command = [
            f'{self.executable}',
            f'{headless_mode}',
            f'--screenshot={os.path.join(output_path, output_file)}',
            f'--window-size={size[0]},{size[1]}',
            *self.flags,
            f'{input}',
        ]

        if self.print_command:
            print(' '.join(command))

        subprocess.run(command, **self._subprocess_run_kwargs)
    
    @property
    def disable_logging(self):
        return self._disable_logging
    
    @disable_logging.setter
    def disable_logging(self, value):
        self._disable_logging = value

        # dict that will be passed unpacked as a parameter
        # to the subprocess.call() method to take a screenshot
        self._subprocess_run_kwargs = {
            'stdout': subprocess.DEVNULL,
            'stderr': subprocess.DEVNULL,
        } if value else {}
    
    def __enter__(self):
        print(
            'Context manager (with ... as:) is',
            f'not supported for {__class__.__name__}.'
        )

    def __exit__(self, *exc):
        pass



from .browser import CDPBrowser
from .search_utils import find_chrome

import os
import subprocess

import requests
import json
from websocket import create_connection
import base64

import websocket
# websocket.enableTrace(True) 


class ChromeCDP(CDPBrowser):

    def __init__(
        self, executable=None, flags=None,
        print_command=False, cdp_port=9222,
        disable_logging=False,
    ):
        self.executable = executable
        if not flags:
            # for some reason, default-background-color prevents
            # the browser from running
            self.flags = [
                '--hide-scrollbars',
            ]
        else:
            self.flags = [flags] if isinstance(flags, str) else flags

        self.print_command = print_command
        self.cdp_port = cdp_port
        self._disable_logging = disable_logging

        self._ws = None  # Websocket connection
        self.proc = None  # Headless browser Popen object

        self.__id = 0

    @property
    def executable(self):
        return self._executable

    @executable.setter
    def executable(self, value):
        self._executable = find_chrome(value)

    @property
    def disable_logging(self):
        return self._disable_logging
    
    @disable_logging.setter
    def disable_logging(self, value):
        self._disable_logging = value

    @property
    def ws(self):

        if not self._ws:
            print(f'----------- http://localhost:{self.cdp_port}/json/version')
            r = requests.get(f'http://localhost:{self.cdp_port}/json')  # TODO use page websocket instead of browser one
            print(f'{r.json()}')
            print(f'Using ws url= {r.json()[0]["webSocketDebuggerUrl"]}')
            self._ws = create_connection(r.json()[0]['webSocketDebuggerUrl'])
            print('Successfully connected to ws.')
        return self._ws

    @property
    def _id(self):
        self.__id += 1
        return self.__id

    def cdp_send(self, method, **params):
        """
        """
        print(f'cdp_send: {method} {params}')
        return self.ws.send(
            json.dumps({
                'id': self._id,
                'method': method,
                'params': params,
            })
        )

    def screenshot(
        self,
        input,
        output_path,
        output_file='screenshot.png',
        size=(1920, 1080),
    ):
        """
        """
        # Useful documentation about the Chrome DevTools Protocol:
        # https://chromedevtools.github.io/devtools-protocol/

        # "Enabling" the page allows to receive the Page.loadEventFired event
        self.cdp_send('Page.enable')

        self.cdp_send('Page.navigate', url=input)

        print('wait for page to load')

        # Wait for page to load entirely
        while True:
            message = json.loads(self.ws.recv())
            method = message.get('method')
            print(f'{method}')
            if method == 'Page.loadEventFired':
                break
        
        print('page disable')
        self.cdp_send('Page.disable')

        self.cdp_send(
            'Emulation.setDeviceMetricsOverride',
            width=size[0],
            height=size[1],
            deviceScaleFactor=0,  # 0 disables the override
            mobile=False,
        )

        print('send Page.captureScreenshot')

        self.cdp_send(
            'Page.captureScreenshot',
            # captureBeyondViewport=True,
            # clip={
            #     'width': size[0],
            #     'height': size[1],
            #     'x': 500,
            #     'y': 200,
            #     'scale': 4
            # }
        )

        print('writing to file..')

        # get screenshot data when ready,
        # while potentially skipping unneeded messages
        while True:
            message = json.loads(self.ws.recv())
            # todo capture and display errors ?
            if 'result' in message and 'data' in message['result']:
                # retrive base64 encoded image data
                img_data = message['result']['data']
                break

        # Decode and write image data to file
        with open(os.path.join(output_path, output_file), 'wb') as f:
            f.write(base64.b64decode(img_data))

    def get_page_infos(self):
        """
        """
        self.cdp_send('Page.getLayoutMetrics')

        while True:
            message = json.loads(self.ws.recv())
            print(f'{message}')
            if 'result' in message and 'layoutViewport' in message['result']:
                return message['result']

    def print_pdf(self):
        # TODO : Page.printToPDF
        pass

    def __enter__(self):
        """
        """
        if not self.disable_logging:
            print(
                'Starting headless Chrome with '
                f'--remote-debugging-port={self.cdp_port}.'
            )

        self.flags.append('--remote-allow-origins=*')

        command = [
            f'{self.executable}',
            '--window-size=1920,1080',
            f'--remote-debugging-port={self.cdp_port}',
            '--headless=new',
            '--no-first-run',
            '--no-default-browser-check',
            *self.flags,
        ]

        # if self.print_command:
        if True:
            print(' '.join(command))

        self.proc = subprocess.Popen(command, shell=True)

    def __exit__(self, *exc):
        """
        """
        if self.disable_logging:
            print(f'Closing headless Chrome instance on port {self.cdp_port}.')

        # check if the process is still running
        if self.proc.poll() is None:
            # ensure that it is properly killed
            try:
                self.cdp_send('Browser.close')
                self.ws.close()
                print('(TODO) Closed CDP and WebSocket connections properly.')
            except Exception:
                print('Could not properly close the CDP and WebSocket connections.')
            
            try:
                self.proc.terminate()
                print('Closed Chrome properly.')
            except Exception:
                print('Could not properly kill Chrome.')



# -*- coding: utf-8 -*-
#
import unittest

from websocket._abnf import ABNF, frame_buffer
from websocket._exceptions import WebSocketProtocolException

"""
test_abnf.py
websocket - WebSocket client library for Python

Copyright 2024 engn33r

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class ABNFTest(unittest.TestCase):
    def test_init(self):
        a = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_PING)
        self.assertEqual(a.fin, 0)
        self.assertEqual(a.rsv1, 0)
        self.assertEqual(a.rsv2, 0)
        self.assertEqual(a.rsv3, 0)
        self.assertEqual(a.opcode, 9)
        self.assertEqual(a.data, "")
        a_bad = ABNF(0, 1, 0, 0, opcode=77)
        self.assertEqual(a_bad.rsv1, 1)
        self.assertEqual(a_bad.opcode, 77)

    def test_validate(self):
        a_invalid_ping = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_PING)
        self.assertRaises(
            WebSocketProtocolException,
            a_invalid_ping.validate,
            skip_utf8_validation=False,
        )
        a_bad_rsv_value = ABNF(0, 1, 0, 0, opcode=ABNF.OPCODE_TEXT)
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_rsv_value.validate,
            skip_utf8_validation=False,
        )
        a_bad_opcode = ABNF(0, 0, 0, 0, opcode=77)
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_opcode.validate,
            skip_utf8_validation=False,
        )
        a_bad_close_frame = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, data=b"\x01")
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_close_frame.validate,
            skip_utf8_validation=False,
        )
        a_bad_close_frame_2 = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, data=b"\x01\x8a\xaa\xff\xdd"
        )
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_close_frame_2.validate,
            skip_utf8_validation=False,
        )
        a_bad_close_frame_3 = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_CLOSE, data=b"\x03\xe7"
        )
        self.assertRaises(
            WebSocketProtocolException,
            a_bad_close_frame_3.validate,
            skip_utf8_validation=True,
        )

    def test_mask(self):
        abnf_none_data = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_PING, mask_value=1, data=None
        )
        bytes_val = b"aaaa"
        self.assertEqual(abnf_none_data._get_masked(bytes_val), bytes_val)
        abnf_str_data = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_PING, mask_value=1, data="a"
        )
        self.assertEqual(abnf_str_data._get_masked(bytes_val), b"aaaa\x00")

    def test_format(self):
        abnf_bad_rsv_bits = ABNF(2, 0, 0, 0, opcode=ABNF.OPCODE_TEXT)
        self.assertRaises(ValueError, abnf_bad_rsv_bits.format)
        abnf_bad_opcode = ABNF(0, 0, 0, 0, opcode=5)
        self.assertRaises(ValueError, abnf_bad_opcode.format)
        abnf_length_10 = ABNF(0, 0, 0, 0, opcode=ABNF.OPCODE_TEXT, data="abcdefghij")
        self.assertEqual(b"\x01", abnf_length_10.format()[0].to_bytes(1, "big"))
        self.assertEqual(b"\x8a", abnf_length_10.format()[1].to_bytes(1, "big"))
        self.assertEqual("fin=0 opcode=1 data=abcdefghij", abnf_length_10.__str__())
        abnf_length_20 = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_BINARY, data="abcdefghijabcdefghij"
        )
        self.assertEqual(b"\x02", abnf_length_20.format()[0].to_bytes(1, "big"))
        self.assertEqual(b"\x94", abnf_length_20.format()[1].to_bytes(1, "big"))
        abnf_no_mask = ABNF(
            0, 0, 0, 0, opcode=ABNF.OPCODE_TEXT, mask_value=0, data=b"\x01\x8a\xcc"
        )
        self.assertEqual(b"\x01\x03\x01\x8a\xcc", abnf_no_mask.format())

    def test_frame_buffer(self):
        fb = frame_buffer(0, True)
        self.assertEqual(fb.recv, 0)
        self.assertEqual(fb.skip_utf8_validation, True)
        fb.clear
        self.assertEqual(fb.header, None)
        self.assertEqual(fb.length, None)
        self.assertEqual(fb.mask_value, None)
        self.assertEqual(fb.has_mask(), False)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod


class Browser(ABC):
    """Abstract class representing a web browser."""

    def __init__(self, flags, disable_logging):
        pass

    @property
    @abstractmethod
    def executable(self):
        pass

    @executable.setter
    @abstractmethod
    def executable(self, value):
        pass

    @abstractmethod
    def screenshot(self, *args, **kwargs):
        pass

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, *exc):
        pass

    @property
    @abstractmethod
    def disable_logging(self):
        pass


class CDPBrowser(Browser):
    """A web browser that can be interacted with via Chrome DevTools Protocol.
    """

    def __init__(self, flags, cdp_port, disable_logging):
        pass



from .chromium import ChromiumHeadless
from .search_utils import get_command_origin, find_first_defined_env_var

import subprocess
import platform
import os
import shutil

ENV_VAR_LOOKUP_TOGGLE = 'HTML2IMAGE_TOGGLE_ENV_VAR_LOOKUP'

EDGE_EXECUTABLE_ENV_VAR_CANDIDATES = [
    'HTML2IMAGE_EDGE_BIN',
    'HTML2IMAGE_EDGE_EXE',
    'EDGE_BIN',
    'EDGE_EXE',
]


def _find_edge(user_given_executable=None):
    """ Finds a edge executable.

    Search Edge on a given path. If no path given,
    try to find Edge or Chromium-browser on a Windows or Unix system.

    Parameters
    ----------
    - `user_given_executable`: str (optional)
        + A filepath leading to a Edge executable
        + Or a filename found in the current working directory
        + Or a keyword that executes Edge/ Chromium, ex:
            - 'msedge' on linux and windows systems (typing `start msedge` in a windows cmd works)

    Raises
    ------
    - `FileNotFoundError`
        + If a suitable edge executable could not be found.

    Returns
    -------
    - str
        + Path of the edge executable on the current machine.
    """

    # try to find a edge bin/exe in ENV
    path_from_env = find_first_defined_env_var(
        env_var_list=EDGE_EXECUTABLE_ENV_VAR_CANDIDATES,
        toggle=ENV_VAR_LOOKUP_TOGGLE
    )

    if path_from_env:
        print(
            f'Found a potential edge executable in the {path_from_env} '
            f'environment variable:\n{path_from_env}\n'
        )
        return path_from_env

    # if an executable is given, try to use it
    if user_given_executable is not None:

        # On Windows, we cannot "safely" validate that user_given_executable
        # seems to be a edge executable, as we cannot run it with
        # the --version flag.
        # https://bugs.chromium.org/p/chromium/issues/detail?id=158372
        #
        # We thus do the "bare minimum" and check if user_given_executable
        # is a file, a filepath, or corresponds to a keyword that can be used
        # with the start command, like so: `start user_given_executable`
        if platform.system() == 'Windows':
            command_origin = get_command_origin(user_given_executable)
            if command_origin:
                return command_origin

            # cannot validate user_given_executable
            raise FileNotFoundError()

        # On a non-Windows OS, we can validate in a basic way that
        # user_given_executable leads to a Edge executable,
        # or is a command, using the --version flag
        else:
            try:
                if 'edge' in subprocess.check_output(
                    [user_given_executable, '--version']
                ).decode('utf-8').lower():
                    return user_given_executable
            except Exception:
                pass

        # We got a user_given_executable but couldn't validate it
        raise FileNotFoundError(
            'Failed to find a seemingly valid edge executable '
            'in the given path.'
        )

    # Executable not in ENV or given by the user, try to find it
    # Search for executable on a Windows OS
    if platform.system() == 'Windows':
        prefixes = [
            os.getenv('PROGRAMFILES(X86)'),
            os.getenv('PROGRAMFILES'),
            os.getenv('LOCALAPPDATA'),
        ]

        suffix = "Microsoft\\Edge\\Application\\msedge.exe"

        for prefix in prefixes:
            path_candidate = os.path.join(prefix, suffix)
            if os.path.isfile(path_candidate):
                return path_candidate

    # Search for executable on a Linux OS
    elif platform.system() == "Linux":

        edge_commands = [
            'msedge',
            '/opt/microsoft/msedge/msedge'
        ]

        for edge_command in edge_commands:
            if shutil.which(edge_command):
                # check the --version for "edge" ?
                return edge_command

    # Search for executable on MacOS
    elif platform.system() == "Darwin":
        # MacOS system
        edge_app = (
            '/Applications/Microsoft Edge.app/Contents/MacOS/Microsoft Edge'
        )

        try:
            version_result = subprocess.check_output(
                [edge_app, "--version"]
            )
            if "Microsoft Edge" in str(version_result):
                return edge_app
        except Exception:
            pass

    # Couldn't find an executable (or OS not in Windows, Linux or Mac)
    raise FileNotFoundError(
        'Could not find a Edge executable on this '
        'machine, please specify it yourself.'
    )

class EdgeHeadless(ChromiumHeadless):
    """
        Edge browser wrapper.

        Parameters
        ----------
        - `executable` : str, optional
            + Path to a edge executable.
        - `flags` : list of str
            + Flags to be used by the headless browser.
            + Default flags are :
                - '--default-background-color=00000000'
                - '--hide-scrollbars'
        - `print_command` : bool
            + Whether or not to print the command used to take a screenshot.
        - `disable_logging` : bool
            + Whether or not to disable Chrome's output.
        - `use_new_headless` : bool, optional
            + Whether or not to use the new headless mode.
            + By default, the old headless mode is used.
            + You can also keep the original behavior to backward compatibility by setting this to `None`.
    """

    def __init__(self, executable=None, flags=None, print_command=False, disable_logging=False, use_new_headless=None,):
        super().__init__(executable=executable, flags=flags, print_command=print_command, disable_logging=disable_logging, use_new_headless=use_new_headless)

    @property
    def executable(self):
        return self._executable

    @executable.setter
    def executable(self, value):
        self._executable = _find_edge(value)



from .browser import Browser
from .search_utils import find_firefox

import os
import subprocess

import requests
import json
from websocket import create_connection
import base64


class FirefoxCDP(Browser):
    def __init__(self, executable=None, flags=None, print_command=False):
        self.executable = executable
        if not flags:
            self.flags = [
                # '--default-background-color=0',
                # '--hide-scrollbars',
            ]
        else:
            self.flags = [flags] if isinstance(flags, str) else flags

        self.print_command = print_command

        self.ws = None  # Websocket connection
        self.proc = None  # Headless browser Popen object

        self.__id = 0

    @property
    def executable(self):
        return self._executable

    @executable.setter
    def executable(self, value):
        self._executable = find_firefox(value)

    @property
    def _id(self):
        self.__id += 1
        return self.__id

    def cdp_send(self, method, **params):
        """
        """
        return self.ws.send(
            json.dumps({
                'id': self._id,
                'method': method,
                'params': params,
            })
        )

    def screenshot(
        self,
        input,
        output_path,
        output_file='screenshot.png',
        size=(1920, 1080),
    ):
        """
        """
        # Useful documentation about the Chrome DevTools Protocol:
        # https://chromedevtools.github.io/devtools-protocol/

        # enabling the page allow to receive the Page.loadEventFired event
        self.cdp_send('Page.enable')
        print(self.ws.recv())
        
        self.cdp_send('Page.navigate', url=input)
        print('Navigated to:', input)

        # wait for page to load entirely
        while True:
            print('Waiting for page to load...')
            message = json.loads(self.ws.recv())
            print(message)
            if message.get('method') == 'Page.loadEventFired':
                break
        
        print('Page loaded.')

        self.cdp_send('Page.disable')
        self.cdp_send(
            'Emulation.setDeviceMetricsOverride',
            width=size[0],
            height=size[1],
            deviceScaleFactor=0,  # 0 disables the override
            mobile=False,
        )
        self.cdp_send('Page.captureScreenshot')

        # get screenshot data when ready,
        # while potentially skipping unneeded messages
        while True:
            message = json.loads(self.ws.recv())
            if 'result' in message and 'data' in message['result']:
                # retrive base64 encoded image data
                img_data = message['result']['data']
                break

        # decode and write image data to file
        with open(os.path.join(output_path, output_file), 'wb') as f:
            f.write(base64.b64decode(img_data))

    def __enter__(self):
        """
        """
        print('Entering context. (ChromeHeadlessServer)')

        temp_dir = os.environ['TMP'] if os.name == 'nt' else '/tmp'
        temp_dir = os.path.join(temp_dir, 'firefox-html2image-profile')

        # create a user.js file that overrides Firefox's default config
        # http://kb.mozillazine.org/User.js_file
        os.makedirs(temp_dir, exist_ok=True)
        with open(os.path.join(temp_dir, 'user.js'), 'w') as f:
            f.write(
                '// File generated by html2image\n'
                'user_pref("devtools.chrome.enabled", true);\n'
                'user_pref("devtools.debugger.prompt-connection", false);\n'
                'user_pref("devtools.debugger.remote-enabled", true);\n'
                '\n'
                '// https://bugzilla.mozilla.org/show_bug.cgi?id=1746154#c4 \n'
                'user_pref("fission.bfcacheInParent", false);\n'
                'user_pref("fission.webContentIsolationStrategy", 0);\n'
            )

        command = [
            f'{self.executable}',
            # '--headless',
            *self.flags,
            '--profile', temp_dir,
            '--new-instance',
            '--remote-debugging-port=9223',
            # 'about:home',
            # browser.newtabpage.activity-stream.asrouter.providers.onboarding
        ]

        if self.print_command or True:
            print(' '.join(command))

        self.proc = subprocess.Popen(command)

        r = requests.get('http://localhost:9223/json/list')

        self.ws = create_connection(
            r.json()[0]['webSocketDebuggerUrl'],

            # Firefox 97.0 requires an empty origin header
            suppress_origin=True,
        )
        # except ConnectionRefusedError as e:
        #     # devtools.chrome.enabled true
        #     # devtools.debugger.prompt-connection false
        #     # devtools.debugger.remote-enabled true
        #     print('Could not connect to Firefox cdp instance.')

    def __exit__(self, *exc):
        """
        """
        print(f'Exiting context. (ChromeHeadlessServer), {exc}')

        self.cdp_send('Browser.close')
        self.ws.close()
        self.proc.terminate()  # ensure that the process is terminated



""" html2image Command Line Interface
"""

import argparse
import os
from html2image import Html2Image


def size_type(string):
    try:
        width, height = map(int, string.split(','))
        if width <= 0 or height <= 0:
            raise argparse.ArgumentTypeError(
                'Width and height must be positive integers.'
            )
        return width, height
    except ValueError:  # incorrect number of values
        raise argparse.ArgumentTypeError(
            f"Size should be W,H (e.g., 1920,1080), instead got '{string}'"
        )
    except Exception as e:  # unexpected errors
        raise argparse.ArgumentTypeError(f"Invalid size format '{string}': {e}")


def main():
    parser = argparse.ArgumentParser(
        description='Generate images from HTML/CSS or URLs using the html2image library.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Html2Image instantiation arguments
    group_hti_init = parser.add_argument_group('Html2Image Instance Configuration')
    group_hti_init.add_argument(
        '--output-path', '-o',
        default=os.getcwd(),
        help='Directory to save screenshots.'
    )

    # TODO : this list is duplicated from browser_map in html2image.py
    browser_choices = [
        'chrome', 'chromium', 'google-chrome', 'google-chrome-stable',
        'googlechrome', 'edge', 'chrome-cdp', 'chromium-cdp'
    ]
    group_hti_init.add_argument(
        '--browser',
        default='chrome',
        choices=browser_choices,
        help='Browser to use for screenshots.'
    )
    group_hti_init.add_argument(
        '--browser-executable',
        default=None,
        help='Path to the browser executable. Auto-detected if not provided.'
    )
    group_hti_init.add_argument(
        '--cdp-port',
        type=int,
        default=None,
        help='CDP port for CDP-enabled browsers (e.g., chrome-cdp). Default is library-dependent.'
    )
    group_hti_init.add_argument(
        '--temp-path',
        default=None,
        help="Directory for temporary files. Defaults to system temp directory within an 'html2image' subfolder."
    )
    group_hti_init.add_argument(
        '--keep-temp-files',
        action='store_true',
        help='Do not delete temporary files after screenshot generation.'
    )
    group_hti_init.add_argument(
        '--custom-flags',
        nargs='*', 
        default=[],  # If not provided, defaults are used
        help="Custom flags to pass to the browser (e.g., '--no-sandbox' '--disable-gpu'). If provided, these flags will be used."
    )

    # Screenshot sources arguments
    group_sources = parser.add_argument_group('Screenshot Sources (at least one type is required)')
    group_sources.add_argument(
        '--url', '-U',
        nargs='*', default=[],
        metavar='URL',
        help='URL(s) to screenshot.'
    )
    group_sources.add_argument(
        '--html-file',
        nargs='*', default=[],
        metavar='FILE',
        help='HTML file(s) to screenshot.'
    )
    group_sources.add_argument(
        '--html-string',
        nargs='*', default=[],
        metavar='STRING',
        help='HTML string(s) to screenshot.'
    )
    group_sources.add_argument(
        '--css-file',
        nargs='*', default=[],
        metavar='FILE',
        help='CSS file(s) to load. Used by HTML files or applied with HTML strings.'
    )
    group_sources.add_argument(
        '--css-string',
        nargs='*', default=[],
        metavar='STRING',
        help='CSS string(s) to apply. Combined and used with HTML strings.'
    )
    group_sources.add_argument(
        '--other-file', '-O',
        nargs='*', default=[],
        metavar='FILE',
        help='Other file(s) to screenshot (e.g., SVG).'
    )

    # Screenshot output control arguments
    group_output_ctrl = parser.add_argument_group('Screenshot Output Options')
    group_output_ctrl.add_argument(
        '--save-as', '-S',
        nargs='*', default=None,  # html2image handles default naming if only one source
        metavar='FILENAME',
        help='Filename(s) for the output images. If not provided or fewer names than items, names are auto-generated.'
    )
    group_output_ctrl.add_argument(
        '--size', '-s',
        nargs='*', default=[],
        type=size_type,
        metavar='W,H',
        help="Size(s) for screenshots as W,H. If one W,H pair is given, it applies to all. If multiple, they apply to corresponding screenshots; if fewer pairs than items, the last is repeated. If omitted, (1920,1080) is used."
    )

    # General arguments
    group_general = parser.add_argument_group('General Options')
    group_general.add_argument(
        '--quiet', '-q',
        action='store_true',
        help='Suppress output from browsers (sets disable_logging=True).'
    )
    group_general.add_argument(
        '--verbose', '-v',
        action='store_true',
        help='Enable verbose output, including browser commands if supported by the browser handler.'
    )

    args = parser.parse_args()

    # Prepare Html2Image()
    hti_kwargs = {
        'output_path': args.output_path,
        'browser': args.browser,
        'browser_executable': args.browser_executable,
        'custom_flags': [cf.replace("'", '') for cf in args.custom_flags],
        'disable_logging': args.quiet,
        'temp_path': args.temp_path,
        'keep_temp_files': args.keep_temp_files,
    }

    # Only pass cdp_port if a CDP browser is likely selected and port is given
    if args.cdp_port and 'cdp' in args.browser.lower():
        hti_kwargs['browser_cdp_port'] = args.cdp_port
    elif args.cdp_port:
        print(
            f"Warning: --cdp-port ({args.cdp_port}) was specified, but the selected browser ('{args.browser}') might not be a CDP browser."
        )

    try:
        # Filter out None values so defaults are used for those specific kwargs
        # keep_temp_files and disable_logging are bools, always pass them.
        # custom_flags should be passed even if None, so Html2Image can use its defaults or an empty list.
        active_hti_kwargs = {
            k: v for k, v in hti_kwargs.items()
            if v is not None or k in ['keep_temp_files', 'disable_logging', 'custom_flags']
        }
        hti = Html2Image(**active_hti_kwargs)

    except Exception as e:
        print(f'Error: Could not instantiate Html2Image: {e}')
        exit(1)

    if args.verbose:
        # The `print_command` attribute is specific to ChromiumHeadless.
        # CDP browsers print logs internally.
        if hasattr(hti.browser, 'print_command'):
            hti.browser.print_command = True
            print('Verbose mode: Browser commands will be printed for compatible handlers.')
        else:
            print('Verbose mode enabled. Note: Detailed browser command printing depends on the selected browser handler.')

    has_sources = any([
        args.url, args.html_file, args.html_string, args.other_file
    ])

    # Print help message if no sources were passed
    if not has_sources:
        print('Error: No screenshot sources (URL, HTML file/string, other file) provided.')
        parser.print_usage()
        exit(1)

    # Perform screenshot
    screenshot_kwargs = {
        'url': args.url,
        'html_file': args.html_file,
        'html_str': args.html_string,
        'css_file': args.css_file,
        'css_str': args.css_string,
        'other_file': args.other_file,
        'size': args.size,  # Pass the list of sizes directly from the --size CLI arg
    }

    if args.save_as is not None:
        screenshot_kwargs['save_as'] = args.save_as

    try:
        if args.verbose:
            print('--- Html2Image Instance Configuration ---')
            for k, v in active_hti_kwargs.items():
                print(f'  {k}: {v}')
            print('--- Screenshot Call Arguments ---')
            for k, v in screenshot_kwargs.items():
                if v or k == 'size':  # print if list not empty, or always for size
                    print(f'  {k}: {v}')

        paths = hti.screenshot(**screenshot_kwargs)

        if not args.quiet:
            print(f'Successfully created {len(paths)} image(s):')
            for path in paths:
                print(f'  {path}')

    except FileNotFoundError as e:
        print(f'Error: A required file was not found: {e}')
        exit(1)

    except ValueError as e:  # Can be raised by browser screenshot method for bad size etc.
        print(f'Error: Invalid value encountered: {e}')
        exit(1)

    except Exception as e:
        print(f'An unexpected error occurred during screenshotting: {e}')
        if args.verbose:
            import traceback
            traceback.print_exc()
        exit(1)


if __name__ == '__main__':
    main()
